##########################################################################
#
#  Copyright (c) 2014, Image Engine Design Inc. All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are
#  met:
#
#      * Redistributions of source code must retain the above
#        copyright notice, this list of conditions and the following
#        disclaimer.
#
#      * Redistributions in binary form must reproduce the above
#        copyright notice, this list of conditions and the following
#        disclaimer in the documentation and/or other materials provided with
#        the distribution.
#
#      * Neither the name of John Haddon nor the names of
#        any other contributors to this software may be used to endorse or
#        promote products derived from this software without specific prior
#        written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
#  IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
#  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
#  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
#  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
#  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
#  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
#  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
#  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
##########################################################################

import os
import unittest

import imath

import IECore
import IECoreScene
import IECoreGL

import Gaffer
import GafferTest
import GafferUI
import GafferUITest
import GafferScene
import GafferSceneUI

class SceneGadgetTest( GafferUITest.TestCase ) :

	def testBound( self ) :

		s = Gaffer.ScriptNode()
		s["p"] = GafferScene.Plane()
		s["g"] = GafferScene.Group()
		s["g"]["in"][0].setInput( s["p"]["out"] )
		s["g"]["transform"]["translate"]["x"].setValue( 2 )

		sg = GafferSceneUI.SceneGadget()
		sg.setScene( s["g"]["out"] )

		sg.waitForCompletion()
		self.assertEqual( sg.bound(), s["g"]["out"].bound( "/" ) )

		s["g"]["transform"]["translate"]["y"].setValue( 4 )
		sg.waitForCompletion()
		self.assertEqual( sg.bound(), s["g"]["out"].bound( "/" ) )

	def assertObjectAt( self, gadget, ndcPosition, path ) :

		viewportGadget = gadget.ancestor( GafferUI.ViewportGadget )

		rasterPosition = ndcPosition * imath.V2f( viewportGadget.getViewport() )
		gadgetLine = viewportGadget.rasterToGadgetSpace( rasterPosition, gadget )

		self.assertEqual( gadget.objectAt( gadgetLine ), path )

	def assertObjectsAt( self, gadget, ndcBox, paths ) :

		viewportGadget = gadget.ancestor( GafferUI.ViewportGadget )

		rasterMin = ndcBox.min() * imath.V2f( viewportGadget.getViewport() )
		rasterMax = ndcBox.max() * imath.V2f( viewportGadget.getViewport() )

		gadgetMin = viewportGadget.rasterToGadgetSpace( rasterMin, gadget ).p0
		gadgetMax = viewportGadget.rasterToGadgetSpace( rasterMax, gadget ).p1

		objectsAt = IECore.PathMatcher()
		gadget.objectsAt( gadgetMin, gadgetMax, objectsAt )

		objects = set( objectsAt.paths() )
		expectedObjects = set( IECore.PathMatcher( paths ).paths() )
		self.assertEqual( objects, expectedObjects )

	def testObjectVisibility( self ) :

		s = Gaffer.ScriptNode()
		s["s"] = GafferScene.Sphere()
		s["g"] = GafferScene.Group()
		s["g"]["in"][0].setInput( s["s"]["out"] )
		s["a"] = GafferScene.StandardAttributes()
		s["a"]["in"].setInput( s["g"]["out"] )

		sg = GafferSceneUI.SceneGadget()
		sg.setMinimumExpansionDepth( 1 )
		sg.setScene( s["a"]["out"] )

		with GafferUI.Window() as w :
			gw = GafferUI.GadgetWidget( sg )

		w.setVisible( True )
		self.waitForIdle( 1000 )

		sg.waitForCompletion()
		gw.getViewportGadget().frame( sg.bound() )

		self.assertObjectAt( sg, imath.V2f( 0.5 ), IECore.InternedStringVectorData( [ "group", "sphere" ] ) )

		s["a"]["attributes"]["visibility"]["enabled"].setValue( True )
		s["a"]["attributes"]["visibility"]["value"].setValue( False )

		sg.waitForCompletion()
		self.assertObjectAt( sg, imath.V2f( 0.5 ), None )

		s["a"]["attributes"]["visibility"]["enabled"].setValue( True )
		s["a"]["attributes"]["visibility"]["value"].setValue( True )

		sg.waitForCompletion()
		self.assertObjectAt( sg, imath.V2f( 0.5 ), IECore.InternedStringVectorData( [ "group", "sphere" ] ) )

	@unittest.skipIf( "TF_BUILD" in os.environ, "Unknown problem running on Azure Pipelines" )
	def testExpansion( self ) :

		s = Gaffer.ScriptNode()
		s["s"] = GafferScene.Sphere()
		s["g"] = GafferScene.Group()
		s["g"]["in"][0].setInput( s["s"]["out"] )
		s["a"] = GafferScene.StandardAttributes()
		s["a"]["in"].setInput( s["g"]["out"] )

		sg = GafferSceneUI.SceneGadget()
		sg.setScene( s["a"]["out"] )

		with GafferUI.Window() as w :
			gw = GafferUI.GadgetWidget( sg )

		w.setVisible( True )
		self.waitForIdle( 10000 )

		sg.waitForCompletion()
		gw.getViewportGadget().frame( sg.bound() )
		self.waitForIdle( 10000 )

		self.assertObjectAt( sg, imath.V2f( 0.5 ), None )
		self.assertObjectsAt( sg, imath.Box2f( imath.V2f( 0 ), imath.V2f( 1 ) ), [ "/group" ] )

		sg.setExpandedPaths( IECore.PathMatcher( [ "/group" ] ) )
		sg.waitForCompletion()

		self.assertObjectAt( sg, imath.V2f( 0.5 ), IECore.InternedStringVectorData( [ "group", "sphere" ] ) )
		self.assertObjectsAt( sg, imath.Box2f( imath.V2f( 0 ), imath.V2f( 1 ) ), [ "/group/sphere" ] )

		sg.setExpandedPaths( IECore.PathMatcher( [] ) )
		sg.waitForCompletion()

		self.assertObjectAt( sg, imath.V2f( 0.5 ), None )
		self.assertObjectsAt( sg, imath.Box2f( imath.V2f( 0 ), imath.V2f( 1 ) ), [ "/group" ] )

	def testExpressions( self ) :

		s = Gaffer.ScriptNode()
		s["p"] = GafferScene.Plane()
		s["g"] = GafferScene.Group()
		s["g"]["in"][0].setInput( s["p"]["out"] )
		s["g"]["in"][1].setInput( s["p"]["out"] )
		s["g"]["in"][2].setInput( s["p"]["out"] )

		s["e"] = Gaffer.Expression()
		s["e"].setExpression( "parent['p']['dimensions']['x'] = 1 + context.getFrame() * 0.1" )

		g = GafferSceneUI.SceneGadget()
		g.setScene( s["g"]["out"] )
		g.bound()

	def testGLResourceDestruction( self ) :

		s = Gaffer.ScriptNode()
		s["p"] = GafferScene.Plane()
		s["g"] = GafferScene.Group()
		s["g"]["in"][0].setInput( s["p"]["out"] )
		s["g"]["in"][1].setInput( s["p"]["out"] )
		s["g"]["in"][2].setInput( s["p"]["out"] )
		s["g"]["in"][3].setInput( s["p"]["out"] )

		sg = GafferSceneUI.SceneGadget()
		sg.setScene( s["g"]["out"] )
		sg.setMinimumExpansionDepth( 2 )

		with GafferUI.Window() as w :
			gw = GafferUI.GadgetWidget( sg )
		w.setVisible( True )

		# Reduce the GL cache size so that not everything will fit, and we'll
		# need to dispose of some objects. We can't dispose of objects on any
		# old thread, just the main GL thread, so it's important that we test
		# that we're doing that appropriately.
		IECoreGL.CachedConverter.defaultCachedConverter().setMaxMemory( 100 )

		for i in range( 1, 1000 ) :
			s["p"]["dimensions"]["x"].setValue( i )
			self.waitForIdle( 10 )

	def testExceptionsDuringCompute( self ) :

		# Make this scene
		#
		# - bigSphere
		#	- littleSphere (with exception in attributes expression)

		s = Gaffer.ScriptNode()

		s["s1"] = GafferScene.Sphere()
		s["s1"]["name"].setValue( "bigSphere" )

		s["s2"] = GafferScene.Sphere()
		s["s2"]["name"].setValue( "littleSphere" )
		s["s2"]["radius"].setValue( 0.1 )

		s["p"] = GafferScene.Parent()
		s["p"]["in"].setInput( s["s1"]["out"] )
		s["p"]["child"].setInput( s["s2"]["out"] )
		s["p"]["parent"].setValue( "/bigSphere" )

		s["a"] = GafferScene.StandardAttributes()
		s["a"]["in"].setInput( s["p"]["out"] )
		s["a"]["attributes"]["doubleSided"]["enabled"].setValue( True )

		s["e"] = Gaffer.Expression()
		s["e"].setExpression( 'parent["a"]["attributes"]["doubleSided"]["value"] = context["nonexistent"]' )

		s["f"] = GafferScene.PathFilter()
		s["f"]["paths"].setValue( IECore.StringVectorData( [ "/bigSphere/littleSphere" ] ) )

		s["a"]["filter"].setInput( s["f"]["out"] )

		# Try to view it

		sg = GafferSceneUI.SceneGadget()
		sg.setScene( s["a"]["out"] )
		sg.setMinimumExpansionDepth( 4 )

		with GafferUI.Window() as w :
			gw = GafferUI.GadgetWidget( sg )
			gw.getViewportGadget().setPlanarMovement( False )
			gw.getViewportGadget().setCamera(
				IECoreScene.Camera( parameters = { "projection" : "perspective", } )
			)

		originalMessageHandler = IECore.MessageHandler.getDefaultHandler()
		mh = IECore.CapturingMessageHandler()
		IECore.MessageHandler.setDefaultHandler(
			IECore.LevelFilteredMessageHandler( mh, IECore.LevelFilteredMessageHandler.defaultLevel() )
		)

		try :

			w.setVisible( True )
			self.waitForIdle( 1000 )
			sg.waitForCompletion()

			# Check we were told about the problem

			self.assertEqual( len( mh.messages ), 1 )
			self.assertEqual( mh.messages[0].level, mh.Level.Error )
			self.assertTrue( "nonexistent" in mh.messages[0].message )

			# And that there isn't some half-assed partial scene
			# being displayed.

			self.assertTrue( sg.bound().isEmpty() )
			gw.getViewportGadget().frame( imath.Box3f( imath.V3f( -1 ), imath.V3f( 1 ) ) )
			self.assertObjectAt( sg, imath.V2f( 0.5 ), None )

			# And that redraws don't cause more fruitless attempts
			# to compute the scene.

			gw.getViewportGadget().frame( imath.Box3f( imath.V3f( -1.1 ), imath.V3f( 1.1 ) ) )
			self.waitForIdle( 1000 )

			self.assertEqual( len( mh.messages ), 1 )
			self.assertObjectAt( sg, imath.V2f( 0.5 ), None )
			self.assertTrue( sg.bound().isEmpty() )

			# Fix the problem with the scene, and check that we can see something now

			s["f"]["enabled"].setValue( False )
			sg.waitForCompletion()

			self.assertEqual( len( mh.messages ), 1 )
			self.assertFalse( sg.bound().isEmpty() )
			self.assertObjectAt( sg, imath.V2f( 0.5 ), IECore.InternedStringVectorData( [ "bigSphere" ] ) )

		finally :

			IECore.MessageHandler.setDefaultHandler( originalMessageHandler )

	def testObjectsAt( self ) :

		plane = GafferScene.Plane()

		sphere = GafferScene.Sphere()
		sphere["radius"].setValue( 0.25 )

		instancer = GafferScene.Instancer()
		instancer["in"].setInput( plane["out"] )
		instancer["instances"].setInput( sphere["out"] )
		instancer["parent"].setValue( "/plane" )

		subTree = GafferScene.SubTree()
		subTree["in"].setInput( instancer["out"] )
		subTree["root"].setValue( "/plane" )

		sg = GafferSceneUI.SceneGadget()
		sg.setScene( subTree["out"] )
		sg.setMinimumExpansionDepth( 100 )

		with GafferUI.Window() as w :
			gw = GafferUI.GadgetWidget( sg )
		w.setVisible( True )
		self.waitForIdle( 10000 )

		gw.getViewportGadget().frame( sg.bound() )
		self.waitForIdle( 10000 )

		self.assertObjectsAt(
			sg,
			imath.Box2f( imath.V2f( 0 ), imath.V2f( 1 ) ),
			[ "/instances/sphere/{}".format( i ) for i in range( 0, 4 ) ]
		)

		self.assertObjectsAt(
			sg,
			imath.Box2f( imath.V2f( 0 ), imath.V2f( 0.5 ) ),
			[ "/instances/sphere/2" ]
		)

		self.assertObjectsAt(
			sg,
			imath.Box2f( imath.V2f( 0.5, 0 ), imath.V2f( 1, 0.5 ) ),
			[ "/instances/sphere/3" ]
		)

		self.assertObjectsAt(
			sg,
			imath.Box2f( imath.V2f( 0, 0.5 ), imath.V2f( 0.5, 1 ) ),
			[ "/instances/sphere/0" ]
		)

		self.assertObjectsAt(
			sg,
			imath.Box2f( imath.V2f( 0.5 ), imath.V2f( 1 ) ),
			[ "/instances/sphere/1" ]
		)

	def testSetAndGetScene( self ) :

		plane = GafferScene.Plane()
		sphere = GafferScene.Sphere()

		sg = GafferSceneUI.SceneGadget()
		self.assertEqual( sg.getScene(), None )

		sg.setScene( plane["out"] )
		self.assertEqual( sg.getScene(), plane["out"] )

		sg.setScene( sphere["out"] )
		self.assertEqual( sg.getScene(), sphere["out"] )

	def testBoundOfUnexpandedEmptyChildren( self ) :

		group1 = GafferScene.Group()
		group2 = GafferScene.Group()
		group2["in"][0].setInput( group1["out"] )

		sg = GafferSceneUI.SceneGadget()
		sg.setScene( group2["out"] )

		sg.waitForCompletion()
		self.assertEqual( sg.bound(), imath.Box3f() )

	def testSelectionMaskAccessors( self ) :

		sg = GafferSceneUI.SceneGadget()
		self.assertEqual( sg.getSelectionMask(), None )

		m = IECore.StringVectorData( [ "MeshPrimitive" ] )
		sg.setSelectionMask( m )
		self.assertEqual( sg.getSelectionMask(), m )

		m.append( "Camera" )
		self.assertNotEqual( sg.getSelectionMask(), m )
		sg.setSelectionMask( m )
		self.assertEqual( sg.getSelectionMask(), m )

		sg.setSelectionMask( None )
		self.assertEqual( sg.getSelectionMask(), None )

	def testSelectionMask( self ) :

		plane = GafferScene.Plane()
		plane["dimensions"].setValue( imath.V2f( 10 ) )
		plane["transform"]["translate"]["z"].setValue( 4 )

		camera = GafferScene.Camera()
		group = GafferScene.Group()
		group["in"][0].setInput( plane["out"] )
		group["in"][1].setInput( camera["out"] )

		sg = GafferSceneUI.SceneGadget()
		sg.setScene( group["out"] )
		sg.setMinimumExpansionDepth( 100 )

		with GafferUI.Window() as w :
			gw = GafferUI.GadgetWidget( sg )
		w.setVisible( True )
		self.waitForIdle( 10000 )

		sg.waitForCompletion()
		gw.getViewportGadget().frame( sg.bound(), imath.V3f( 0, 0, -1 ) )
		self.waitForIdle( 10000 )

		self.assertObjectsAt(
			sg,
			imath.Box2f( imath.V2f( 0 ), imath.V2f( 1 ) ),
			[ "/group/plane", "/group/camera" ]
		)

		sg.setSelectionMask( IECore.StringVectorData( [ "MeshPrimitive" ] ) )

		self.assertObjectsAt(
			sg,
			imath.Box2f( imath.V2f( 0 ), imath.V2f( 1 ) ),
			[ "/group/plane" ]
		)

		sg.setSelectionMask( IECore.StringVectorData( [ "Camera" ] ) )

		self.assertObjectsAt(
			sg,
			imath.Box2f( imath.V2f( 0 ), imath.V2f( 1 ) ),
			[ "/group/camera" ]
		)

	def setUp( self ) :

		GafferUITest.TestCase.setUp( self )

		self.__cachedConverterMaxMemory = IECoreGL.CachedConverter.defaultCachedConverter().getMaxMemory()

	def tearDown( self ) :

		GafferUITest.TestCase.tearDown( self )

		IECoreGL.CachedConverter.defaultCachedConverter().setMaxMemory( self.__cachedConverterMaxMemory )

if __name__ == "__main__":
	unittest.main()
